#include <SPI.h>
#include <GD.h>

static PROGMEM prog_uint32_t crc_table[16] = {
    0x00000000, 0x1db71064, 0x3b6e20c8, 0x26d930ac,
    0x76dc4190, 0x6b6b51f4, 0x4db26158, 0x5005713c,
    0xedb88320, 0xf00f9344, 0xd6d6a3e8, 0xcb61b38c,
    0x9b64c2b0, 0x86d3d2d4, 0xa00ae278, 0xbdbdf21c
};

unsigned long crc_update(unsigned long crc, byte data)
{
    byte tbl_idx;
    tbl_idx = crc ^ (data >> (0 * 4));
    crc = pgm_read_dword_near(crc_table + (tbl_idx & 0x0f)) ^ (crc >> 4);
    tbl_idx = crc ^ (data >> (1 * 4));
    crc = pgm_read_dword_near(crc_table + (tbl_idx & 0x0f)) ^ (crc >> 4);

    return crc;
}

#define FLOCAL 2
#define FREMOTE 10

#define SEL_local() digitalWrite(FLOCAL, LOW)
#define UNSEL_local() digitalWrite(FLOCAL, HIGH)
#define SEL_remote() digitalWrite(FREMOTE, LOW)
#define UNSEL_remote() digitalWrite(FREMOTE, HIGH)

#define spix(n) SPI.transfer(n)

static void spipage(int n)
{
  spix(n >> 7);
  spix(n << 1);
  spix(0);
}

static unsigned long pagecrc(int n)
{
  SPI.transfer(0x03);
  spipage(n);

  unsigned long crc = ~0;
  int len = 264;
  while (len--) {
    crc = crc_update(crc, SPI.transfer(0));
  }
  crc = ~crc;
  return crc;
}

static unsigned long local_pagecrc(int n)
{
  SEL_local();
  unsigned long r = pagecrc(n);
  UNSEL_local();
  return r;
}

static unsigned long remote_pagecrc(int n)
{
  SEL_remote();
  unsigned long r = pagecrc(n);
  UNSEL_remote();
  return r;
}

static byte remote_status()
{
  SEL_remote();
  spix(0xd7);   // read SPI flash status
  byte status = spix(0);
  UNSEL_remote();
  return status;
}

static byte img[264];

static unsigned long clone(int n)
{
  int i;

  SEL_local();
  SPI.transfer(0x03);
  spipage(n);
  for (i = 0; i < 264; i++)
    img[i] = spix(0);
  UNSEL_local();

  SEL_remote();
  SPI.transfer(0x82);
  spipage(n);
  for (i = 0; i < 264; i++)
    spix(img[i]);
  UNSEL_remote();
  while ((remote_status() & 0x80) == 0)
    ;

  delay(10);
}

#define PAGES 600   // only first part of flash matters

void setup()
{
  Serial.begin(1000000);

  GD.begin();

  GD.wr(IOMODE, 'F');

  pinMode(FLOCAL, OUTPUT);
  pinMode(FREMOTE, OUTPUT);

  UNSEL_local();
  UNSEL_remote();
  delay(100);
#if 0
  int p;
  long started = millis();
  if (0)
    for (p = 0; p < PAGES; p++)
      clone(p);
  Serial.println(verify(), DEC);
  Serial.println(millis() - started, DEC);
#endif
}

void loop()
{
  char msg[50];

  GD.ascii();
  GD.putstr(0, 0, "Clone utility - factory use only");

  // wait for flash on REMOTE
  byte status;
  do {
    status = remote_status();
    
    sprintf(msg, "New flash status %02x", status);
    GD.putstr(0, 3, msg);
  } while ((status & 0x3c) != 0x14);

  GD.putstr(0, 5, "programming...");
  delay(1000);

  // disable sector protection
  SEL_remote();
  spix(0x3d);
  spix(0x2a);
  spix(0x7f);
  spix(0x9a);
  UNSEL_remote();

  // copy all pages from LOCAL to REMOTE
  int p;
  for (p = 0; p < PAGES; p++) {
    sprintf(msg, "Page %3d", p);
    GD.putstr(0, 7, msg);

    clone(p);
    if (local_pagecrc(p) != remote_pagecrc(p)) {
      GD.putstr(25, 7, "FAILED");
      for (;;) ;
    }
  }
  GD.putstr(0, 9, "SUCCESS");
  // GD.screenshot(0);
  for (;;) ;
}
